@@ -27,6 +27,7 @@
 #include "env.h"
 #include "node.h"
 #include "node_binding.h"
+#include "node_file_utils.h"
 #include "node_mutex.h"
 #include "tracing/trace_event.h"
 #include "util.h"
@@ -363,8 +364,8 @@ v8::Isolate* NewIsolate(v8::Isolate::CreateParams* params,
                         const IsolateSettings& settings = {});
 // This overload automatically picks the right 'main_script_id' if no callback
 // was provided by the embedder.
-v8::MaybeLocal<v8::Value> StartExecution(Environment* env,
-                                         StartExecutionCallback cb = nullptr);
+v8::MaybeLocal<v8::Value> StartExecution(
+    Environment* env, StartExecutionCallbackWithModule cb = nullptr);
 v8::MaybeLocal<v8::Object> GetPerContextExports(
     v8::Local<v8::Context> context, IsolateData* isolate_data = nullptr);
 void MarkBootstrapComplete(const v8::FunctionCallbackInfo<v8::Value>& args);
@@ -411,11 +412,6 @@ typedef struct tm TIME_TYPE;
 #endif
 
 double GetCurrentTimeInMicroseconds();
-int WriteFileSync(const char* path, uv_buf_t* bufs, size_t buf_count);
-int WriteFileSync(const char* path, uv_buf_t buf);
-int WriteFileSync(v8::Isolate* isolate,
-                  const char* path,
-                  v8::Local<v8::String> string);
 
 class DiagnosticFilename {
  public:
