@@ -225,7 +225,7 @@ builtins::CodeCacheInfo SnapshotDeserializer::Read() {
 
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<builtins::CodeCacheInfo>() %s\n", str.c_str());
+    Debug("Read<builtins::CodeCacheInfo>() %s\n", str);
   }
   return result;
 }
@@ -234,7 +234,7 @@ template <>
 size_t SnapshotSerializer::Write(const builtins::CodeCacheInfo& info) {
   Debug("\nWrite<builtins::CodeCacheInfo>() id = %s"
         ", length=%d\n",
-        info.id.c_str(),
+        info.id,
         info.data.length);
 
   size_t written_total = WriteString(info.id);
@@ -263,7 +263,7 @@ PropInfo SnapshotDeserializer::Read() {
 
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<PropInfo>() %s\n", str.c_str());
+    Debug("Read<PropInfo>() %s\n", str);
   }
 
   return result;
@@ -273,7 +273,7 @@ template <>
 size_t SnapshotSerializer::Write(const PropInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<PropInfo>() %s\n", str.c_str());
+    Debug("Write<PropInfo>() %s\n", str);
   }
 
   size_t written_total = WriteString(data.name);
@@ -305,7 +305,7 @@ AsyncHooks::SerializeInfo SnapshotDeserializer::Read() {
 
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<AsyncHooks::SerializeInfo>() %s\n", str.c_str());
+    Debug("Read<AsyncHooks::SerializeInfo>() %s\n", str);
   }
 
   return result;
@@ -314,7 +314,7 @@ template <>
 size_t SnapshotSerializer::Write(const AsyncHooks::SerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<AsyncHooks::SerializeInfo>() %s\n", str.c_str());
+    Debug("Write<AsyncHooks::SerializeInfo>() %s\n", str);
   }
 
   size_t written_total =
@@ -341,7 +341,7 @@ TickInfo::SerializeInfo SnapshotDeserializer::Read() {
 
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<TickInfo::SerializeInfo>() %s\n", str.c_str());
+    Debug("Read<TickInfo::SerializeInfo>() %s\n", str);
   }
 
   return result;
@@ -351,7 +351,7 @@ template <>
 size_t SnapshotSerializer::Write(const TickInfo::SerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<TickInfo::SerializeInfo>() %s\n", str.c_str());
+    Debug("Write<TickInfo::SerializeInfo>() %s\n", str);
   }
 
   size_t written_total = WriteArithmetic<AliasedBufferIndex>(data.fields);
@@ -370,7 +370,7 @@ ImmediateInfo::SerializeInfo SnapshotDeserializer::Read() {
   result.fields = ReadArithmetic<AliasedBufferIndex>();
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<ImmediateInfo::SerializeInfo>() %s\n", str.c_str());
+    Debug("Read<ImmediateInfo::SerializeInfo>() %s\n", str);
   }
   return result;
 }
@@ -379,7 +379,7 @@ template <>
 size_t SnapshotSerializer::Write(const ImmediateInfo::SerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<ImmediateInfo::SerializeInfo>() %s\n", str.c_str());
+    Debug("Write<ImmediateInfo::SerializeInfo>() %s\n", str);
   }
 
   size_t written_total = WriteArithmetic<AliasedBufferIndex>(data.fields);
@@ -403,7 +403,7 @@ performance::PerformanceState::SerializeInfo SnapshotDeserializer::Read() {
   result.observers = ReadArithmetic<AliasedBufferIndex>();
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<PerformanceState::SerializeInfo>() %s\n", str.c_str());
+    Debug("Read<PerformanceState::SerializeInfo>() %s\n", str);
   }
   return result;
 }
@@ -413,7 +413,7 @@ size_t SnapshotSerializer::Write(
     const performance::PerformanceState::SerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<PerformanceState::SerializeInfo>() %s\n", str.c_str());
+    Debug("Write<PerformanceState::SerializeInfo>() %s\n", str);
   }
 
   size_t written_total = WriteArithmetic<AliasedBufferIndex>(data.root);
@@ -439,7 +439,7 @@ IsolateDataSerializeInfo SnapshotDeserializer::Read() {
   result.template_values = ReadVector<PropInfo>();
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<IsolateDataSerializeInfo>() %s\n", str.c_str());
+    Debug("Read<IsolateDataSerializeInfo>() %s\n", str);
   }
   return result;
 }
@@ -448,7 +448,7 @@ template <>
 size_t SnapshotSerializer::Write(const IsolateDataSerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("Write<IsolateDataSerializeInfo>() %s\n", str.c_str());
+    Debug("Write<IsolateDataSerializeInfo>() %s\n", str);
   }
 
   size_t written_total = WriteVector<SnapshotIndex>(data.primitive_values);
@@ -473,7 +473,7 @@ template <>
 size_t SnapshotSerializer::Write(const RealmSerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("\nWrite<RealmSerializeInfo>() %s\n", str.c_str());
+    Debug("\nWrite<RealmSerializeInfo>() %s\n", str);
   }
 
   // Use += here to ensure order of evaluation.
@@ -507,7 +507,7 @@ template <>
 size_t SnapshotSerializer::Write(const EnvSerializeInfo& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("\nWrite<EnvSerializeInfo>() %s\n", str.c_str());
+    Debug("\nWrite<EnvSerializeInfo>() %s\n", str);
   }
 
   // Use += here to ensure order of evaluation.
@@ -549,7 +549,7 @@ SnapshotMetadata SnapshotDeserializer::Read() {
 
   if (is_debug) {
     std::string str = ToStr(result);
-    Debug("Read<SnapshotMetadata>() %s\n", str.c_str());
+    Debug("Read<SnapshotMetadata>() %s\n", str);
   }
   return result;
 }
@@ -558,7 +558,7 @@ template <>
 size_t SnapshotSerializer::Write(const SnapshotMetadata& data) {
   if (is_debug) {
     std::string str = ToStr(data);
-    Debug("\nWrite<SnapshotMetadata>() %s\n", str.c_str());
+    Debug("\nWrite<SnapshotMetadata>() %s\n", str);
   }
   size_t written_total = 0;
   // We need the Node.js version, platform and arch to match because
@@ -566,7 +566,7 @@ size_t SnapshotSerializer::Write(const SnapshotMetadata& data) {
   // can be changed in semver-patches.
   Debug("Write snapshot type %d\n", static_cast<uint8_t>(data.type));
   written_total += WriteArithmetic<uint8_t>(static_cast<uint8_t>(data.type));
-  Debug("Write Node.js version %s\n", data.node_version.c_str());
+  Debug("Write Node.js version %s\n", data.node_version);
   written_total += WriteString(data.node_version);
   Debug("Write Node.js arch %s\n", data.node_arch);
   written_total += WriteString(data.node_arch);
@@ -968,6 +968,61 @@ std::optional<SnapshotConfig> ReadSnapshotConfig(const char* config_path) {
   return result;
 }
 
+// Find bindings that have been loaded by internalBinding() but the external
+// reference method have not been called. This requires that the caller
+// match the id passed into their NODE_BINDING_CONTEXT_AWARE_INTERNAL() and
+// NODE_BINDING_EXTERNAL_REFERENCE() calls. Note that this only serves as a
+// preemptive check. Binding methods create the actual external references
+// (usually through function templates) and there's currently no easy way
+// to verify at that level of granularity. See "Registering binding functions
+// used in bootstrap" in src/README.md.
+bool ValidateBindings(Environment* env) {
+  std::set<std::string> registered;
+#define V(modname) registered.insert(#modname);
+  EXTERNAL_REFERENCE_BINDING_LIST(V)
+#undef V
+
+  std::set<std::string> bindings_without_external_references = {
+      "async_context_frame",
+      "constants",
+      "symbols",
+  };
+
+  std::set<std::string> unregistered;
+  for (auto* mod : env->principal_realm()->internal_bindings) {
+    if (registered.count(mod->nm_modname) == 0 &&
+        bindings_without_external_references.count(mod->nm_modname) == 0) {
+      unregistered.insert(mod->nm_modname);
+    }
+  }
+
+  if (unregistered.size() == 0) {
+    return true;
+  }
+
+  FPrintF(
+      stderr,
+      "\n---- snapshot building check failed ---\n\n"
+      "The following bindings are loaded during the snapshot building process,"
+      " but their external reference registration methods have not been "
+      "called:\n\n");
+  for (auto& binding : unregistered) {
+    FPrintF(stderr, " - %s\n", binding);
+  }
+  FPrintF(stderr,
+          "\nIf the binding does not have any external references, "
+          "add it to the list of bindings_without_external_references "
+          "in src/node_snapshotable.cc.\n"
+          "Otherwise, make sure to call NODE_BINDING_EXTERNAL_REFERENCE() "
+          "with an appropriate register method for the binding, "
+          "and add it to EXTERNAL_REFERENCE_BINDING_LIST in "
+          "src/node_external_reference.h"
+          "\n\nSee \"Registering binding functions used in bootstrap\" "
+          "in src/README.md for more details."
+          "\n----\n\n");
+  return false;
+}
+
 ExitCode BuildSnapshotWithoutCodeCache(
     SnapshotData* out,
     const std::vector<std::string>& args,
@@ -1033,6 +1088,11 @@ ExitCode BuildSnapshotWithoutCodeCache(
     if (exit_code != ExitCode::kNoFailure) {
       return exit_code;
     }
+
+    if (snapshot_type == SnapshotMetadata::Type::kDefault &&
+        !ValidateBindings(env)) {
+      return ExitCode::kStartupSnapshotFailure;
+    }
   }
 
   return SnapshotBuilder::CreateSnapshot(out, setup.get());
@@ -1070,8 +1130,8 @@ ExitCode BuildCodeCacheFromSnapshot(SnapshotData* out,
       std::string size_str = FormatSize(item.data.length);
       per_process::Debug(DebugCategory::MKSNAPSHOT,
                          "Generated code cache for %d: %s\n",
-                         item.id.c_str(),
-                         size_str.c_str());
+                         item.id,
+                         size_str);
     }
   }
   return ExitCode::kNoFailure;
